@@ -26,6 +26,7 @@ from zope.interface import implementer
 
 import priority
 import h2.connection
+import h2.errors
 import h2.events
 import h2.exceptions
 
@@ -34,6 +35,8 @@ from twisted.internet.interfaces import (
     IProtocol, ITransport, IConsumer, IPushProducer, ISSLTransport
 )
 from twisted.internet.protocol import Protocol
+from twisted.logger import Logger
+from twisted.protocols.policies import TimeoutMixin
 from twisted.protocols.tls import _PullToPush
 
 
@@ -59,7 +62,7 @@ if sys.version_info < (2, 7, 4):
 
 
 @implementer(IProtocol, IPushProducer)
-class H2Connection(Protocol):
+class H2Connection(Protocol, TimeoutMixin):
     """
     A class representing a single HTTP/2 connection.
 
@@ -104,6 +107,8 @@ class H2Connection(Protocol):
     factory = None
     site = None
 
+    _log = Logger()
+
     def __init__(self, reactor=None):
         self.conn = h2.connection.H2Connection(
             client_side=False, header_encoding=None
@@ -132,6 +137,7 @@ class H2Connection(Protocol):
         by the L{twisted.web.http._GenericHTTPChannelProtocol} during upgrade
         to HTTP/2.
         """
+        self.setTimeout(self.timeOut)
         self.conn.initiate_connection()
         self.transport.write(self.conn.data_to_send())
 
@@ -143,6 +149,8 @@ class H2Connection(Protocol):
         @param data: The data received from the transport.
         @type data: L{bytes}
         """
+        self.resetTimeout()
+
         try:
             events = self.conn.receive_data(data)
         except h2.exceptions.ProtocolError:
@@ -175,6 +183,40 @@ class H2Connection(Protocol):
             self.transport.write(dataToSend)
 
 
+    def timeoutConnection(self):
+        """
+        Called when the connection has been inactive for
+        L{self.timeOut<twisted.protocols.policies.TimeoutMixin.timeOut>}
+        seconds. Cleanly tears the connection down, attempting to notify the
+        peer if needed.
+
+        We override this method to add two extra bits of functionality:
+
+         - We want to log the timeout.
+         - We want to send a GOAWAY frame indicating that the connection is
+           being terminated, and whether it was clean or not. We have to do this
+           before the connection is torn down.
+        """
+        self._log.info(
+            "Timing out client {client}", client=self.transport.getPeer()
+        )
+
+        # Check whether there are open streams. If there are, we're going to
+        # want to use the error code PROTOCOL_ERROR. If there aren't, use
+        # NO_ERROR.
+        if (self.conn.open_outbound_streams > 0 or
+                self.conn.open_inbound_streams > 0):
+            error_code = h2.errors.PROTOCOL_ERROR
+        else:
+            error_code = h2.errors.NO_ERROR
+
+        self.conn.close_connection(error_code=error_code)
+        self.transport.write(self.conn.data_to_send())
+
+        # We're done, throw the connection away.
+        self.transport.loseConnection()
+
+
     def connectionLost(self, reason):
         """
         Called when the transport connection is lost.
@@ -183,6 +225,7 @@ class H2Connection(Protocol):
         lost, and cleans up all internal state.
         """
         self._stillProducing = False
+        self.setTimeout(None)
 
         for stream in self.streams.values():
             stream.connectionLost(reason)
@@ -271,7 +314,8 @@ class H2Connection(Protocol):
     def _sendPrioritisedData(self, *args):
         """
         The data sending loop. This function repeatedly calls itself, either
-        from L{Deferred}s or from L{twisted.internet.reactor.callLater}.
+        from L{Deferred}s or from
+        L{reactor.callLater<twisted.internet.interfaces.IReactorTime.callLater>}
 
         This function sends data on streams according to the rules of HTTP/2
         priority. It ensures that the data from each stream is interleved
@@ -462,8 +506,16 @@ class H2Connection(Protocol):
         @type streamID: L{int}
         """
         headers.insert(0, (b':status', code))
-        self.conn.send_headers(streamID, headers)
-        self.transport.write(self.conn.data_to_send())
+
+        try:
+            self.conn.send_headers(streamID, headers)
+        except h2.exceptions.StreamClosedError:
+            # Stream was closed by the client at some point. We need to not
+            # explode here: just swallow the error. That's what write() does
+            # when a connection is lost, so that's what we do too.
+            return
+        else:
+            self.transport.write(self.conn.data_to_send())
 
 
     def writeDataToStream(self, streamID, data):
@@ -582,7 +634,11 @@ class H2Connection(Protocol):
                 # unnecessary but benign. We'll ignore it.
                 return
 
-            self.priority.unblock(streamID)
+            # If we haven't got any data to send, don't unblock the stream. If
+            # we do, we'll eventually get an exception inside the
+            # _sendPrioritisedData loop some time later.
+            if self._outboundStreamQueues.get(streamID):
+                self.priority.unblock(streamID)
             self.streams[streamID].windowUpdated()
         else:
             # Update strictly applies to all streams.
@@ -629,10 +685,10 @@ class H2Connection(Protocol):
         incremented.
         @type increment: L{int}
         """
-        # TODO: Consider whether we want some kind of consolidating logic here.
-        self.conn.increment_flow_control_window(increment, stream_id=streamID)
-        self.conn.increment_flow_control_window(increment, stream_id=None)
-        self.transport.write(self.conn.data_to_send())
+        self.conn.acknowledge_received_data(increment, streamID)
+        data = self.conn.data_to_send()
+        if data:
+            self.transport.write(data)
 
 
     def _isSecure(self):
@@ -1011,14 +1067,14 @@ class H2Stream(object):
         """
         Get information about the peer.
         """
-        self._conn.getPeer()
+        return self._conn.getPeer()
 
 
     def getHost(self):
         """
         Similar to getPeer, but for this side of the connection.
         """
-        self._conn.getHost()
+        return self._conn.getHost()
 
 
     def isSecure(self):
@@ -1077,7 +1133,7 @@ class H2Stream(object):
 
     def unregisterProducer(self):
         """
-        @see L{IConsumer.unregisterProducer}
+        @see: L{IConsumer.unregisterProducer}
         """
         # When the producer is unregistered, we're done.
         if self.producer is not None and not self.hasStreamingProducer:
@@ -1091,7 +1147,7 @@ class H2Stream(object):
     # Implementation: IPushProducer
     def stopProducing(self):
         """
-        @see L{IProducer.stopProducing}
+        @see: L{IProducer.stopProducing}
         """
         self.producing = False
         self.abortConnection()
@@ -1099,14 +1155,14 @@ class H2Stream(object):
 
     def pauseProducing(self):
         """
-        @see L{IPushProducer.pauseProducing}
+        @see: L{IPushProducer.pauseProducing}
         """
         self.producing = False
 
 
     def resumeProducing(self):
         """
-        @see L{IPushProducer.resumeProducing}
+        @see: L{IPushProducer.resumeProducing}
         """
         self.producing = True
         consumedLength = 0
